VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "SparseArray"
Attribute VB_Creatable = False
Attribute VB_Exposed = False
' ************************************************
' SpArray.CLS
'
' Sparse array class.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

Private RowHead() As SparseArrayCell
Private ColHead() As SparseArrayCell
Public NumRows As Integer
Public NumCols As Integer
Public NoValue As Variant ' Value for empty entries.

' ************************************************
' Return the value of the indicated entry.
' ************************************************
Property Get Value(r As Integer, c As Integer) As Variant
Dim cell As SparseArrayCell

    Value = NoValue ' Assume we will not find it.
    If r < 1 Or c < 1 Or _
        r > NumRows Or c > NumCols _
            Then Exit Property
    
    Set cell = RowHead(r).NextInRow ' Skip sentinel.
    Do
        If cell Is Nothing Then Exit Property  ' Not found.
        If cell.Col > c Then Exit Property     ' Not found.
        If cell.Col = c Then Exit Do           ' Found.
        Set cell = cell.NextInRow
    Loop

    Value = cell.Data
End Property

' ************************************************
' Find the indicated entry. Remove it if it
' exists.
' ************************************************
Private Sub RemoveEntry(r As Integer, c As Integer)
Dim cell As SparseArrayCell
Dim nxt As SparseArrayCell

    If r < 1 Or r > NumRows Or _
       c < 1 Or c > NumCols Then _
            Exit Sub
    
    ' Find it in its row.
    Set cell = RowHead(r)
    Set nxt = cell.NextInRow
    Do
        If nxt Is Nothing Then Exit Do
        If nxt.Col > c Then Exit Do
        If nxt.Col = c Then
            ' Remove the cell from the row list.
            Set cell.NextInRow = nxt.NextInRow
            Exit Do
        End If
        Set cell = nxt
        Set nxt = cell.NextInRow
    Loop
    
    ' Find it in its column.
    Set cell = ColHead(c)
    Set nxt = cell.NextInCol
    Do
        If nxt Is Nothing Then Exit Do
        If nxt.Row > r Then Exit Do
        If nxt.Row = r Then
            ' Remove the cell from the column list.
            Set cell.NextInCol = nxt.NextInCol
            Exit Do
        End If
        Set cell = nxt
        Set nxt = cell.NextInCol
    Loop
End Sub

' ************************************************
' Set the indicated entry, creating a new cell
' if necessary.
' ************************************************
Property Let Value(r As Integer, c As Integer, new_value As Variant)
Dim i As Integer
Dim found_it As Boolean
Dim cell As SparseArrayCell
Dim nxt As SparseArrayCell
Dim new_cell As SparseArrayCell

    ' If value = NoValue, remove the entry from
    ' the array.
    If new_value = NoValue Then
        RemoveEntry r, c
        Exit Property
    End If

    ' Make more rows if needed
    If r > NumRows Then
        ReDim Preserve RowHead(1 To r)

        ' Initialize a sentinel for each new row
        For i = NumRows + 1 To r
            Set RowHead(i) = New SparseArrayCell
        Next i
        NumRows = r
    End If

    ' Make more columns if needed
    If c > NumCols Then
        ReDim Preserve ColHead(1 To c)

        ' Initialize a sentinel for each new row
        For i = NumCols + 1 To c
            Set ColHead(i) = New SparseArrayCell
        Next i
        NumCols = c
    End If

    ' Try to locate the cell in question.
    Set cell = RowHead(r)
    Set nxt = cell.NextInRow
    Do
        If nxt Is Nothing Then Exit Do
        If nxt.Col >= c Then Exit Do
        Set cell = nxt
        Set nxt = cell.NextInRow
    Loop

    ' See if we found it.
    If nxt Is Nothing Then
        found_it = False
    Else
        found_it = (nxt.Col = c)
    End If

    ' If we did not find the cell, make a new one.
    If Not found_it Then
        Set new_cell = New SparseArrayCell

        ' Place the cell in the row.
        Set new_cell.NextInRow = nxt
        Set cell.NextInRow = new_cell

        ' Place the cell in the column list.
        Set cell = ColHead(c)
        Set nxt = cell.NextInCol
        Do
            If nxt Is Nothing Then Exit Do
            If nxt.Col >= c Then Exit Do
            Set cell = nxt
            Set nxt = cell.NextInRow
        Loop

        Set new_cell.NextInCol = nxt
        Set cell.NextInCol = new_cell
        new_cell.Row = r
        new_cell.Col = c

        ' We will put the value in the nxt cell.
        Set nxt = new_cell
    End If

    ' Set the value
    nxt.Data = new_value
End Property
' ************************************************
' Find the indicated cell if it exists. Set
' prev_in_row and prev_in_col to the cells before
' the target cell in its row and column lists.
' ************************************************
Private Sub FindCell(r As Integer, c As Integer, prev_in_row As SparseArrayCell, prev_in_col As SparseArrayCell)
Dim prev As SparseArrayCell
Dim nxt As SparseArrayCell

    ' Locate the cell in its row.
    Set prev = RowHead(r)
    Set nxt = prev.NextInRow
    Do
        If nxt Is Nothing Then Exit Do
        If nxt.Col >= c Then Exit Do
        Set prev = nxt
        Set nxt = prev.NextInRow
    Loop
    Set prev_in_row = prev
    
    ' Locate the cell in its column.
    Set prev = ColHead(r)
    Set nxt = prev.NextInCol
    Do
        If nxt Is Nothing Then Exit Do
        If nxt.Row >= r Then Exit Do
        Set prev = nxt
        Set nxt = prev.NextInCol
    Loop
    Set prev_in_col = prev
End Sub

' ************************************************
' Initialize NoValue.
' ************************************************
Private Sub Class_Initialize()
    NoValue = 0
End Sub

