VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "VerySparseArray"
Attribute VB_Creatable = False
Attribute VB_Exposed = False
' ************************************************
' VSpArray.CLS
'
' Very sparse array class.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

Public RowSentinel As New HeaderCell
Public ColSentinel As New HeaderCell

Public NoValue As Variant ' Value for empty entries.

' ************************************************
' Return the value of the indicated entry.
' ************************************************
Property Get Value(r As Integer, c As Integer) As Variant
Dim prev_hdr As HeaderCell
Dim next_hdr As HeaderCell
Dim prev_cell As VerySparseArrayCell
Dim next_cell As VerySparseArrayCell

    Value = NoValue ' Assume we will not find it.
    
    ' Find the row that should hold the entry.
    Set prev_hdr = RowSentinel
    Do
        Set next_hdr = prev_hdr.NextHeader
        If next_hdr Is Nothing Then Exit Property
        If next_hdr.RowOrCol > r Then Exit Property
        If next_hdr.RowOrCol = r Then Exit Do
        Set prev_hdr = next_hdr
    Loop
    
    ' Locate the cell in this row.
    Set prev_cell = next_hdr.CellSentinel
    Do
        Set next_cell = prev_cell.NextInRow
        If next_cell Is Nothing Then Exit Property
        If next_cell.Col > c Then Exit Property
        If next_cell.Col = c Then Exit Do
        Set prev_cell = next_cell
    Loop

    Value = next_cell.Data
End Property

' ************************************************
' Find the indicated entry and remove it if it
' exists.
' ************************************************
Private Sub RemoveEntry(r As Integer, c As Integer)
Dim prev_hdr As HeaderCell
Dim next_hdr As HeaderCell
Dim prev_cell As VerySparseArrayCell
Dim next_cell As VerySparseArrayCell

    ' ------------------------------
    ' Remove the entry from its row.
    ' ------------------------------
    ' Find the row that should hold the entry.
    Set prev_hdr = RowSentinel
    Do
        Set next_hdr = prev_hdr.NextHeader
        If next_hdr Is Nothing Then Exit Sub
        If next_hdr.RowOrCol > r Then Exit Sub
        If next_hdr.RowOrCol = r Then Exit Do
        Set prev_hdr = next_hdr
    Loop
    
    ' Locate the cell in this row.
    Set prev_cell = next_hdr.CellSentinel
    Do
        Set next_cell = prev_cell.NextInRow
        If next_cell Is Nothing Then Exit Sub
        If next_cell.Col > c Then Exit Sub
        If next_cell.Col = c Then Exit Do
        Set prev_cell = next_cell
    Loop
    
    ' Remove the cell from the row.
    Set prev_cell.NextInRow = next_cell.NextInRow

    ' If this is the only cell in the row,
    ' remove the row.
    If next_hdr.CellSentinel.NextInRow Is Nothing Then
        Set prev_hdr.NextHeader = next_hdr.NextHeader
    End If

    ' ---------------------------------
    ' Remove the entry from its column.
    ' ---------------------------------
    ' Find the column that should hold the entry.
    Set prev_hdr = ColSentinel
    Do
        Set next_hdr = prev_hdr.NextHeader
        If next_hdr.RowOrCol = c Then Exit Do
        Set prev_hdr = next_hdr
    Loop
    
    ' Locate the cell in this column.
    Set prev_cell = next_hdr.CellSentinel
    Do
        Set next_cell = prev_cell.NextInCol
        If next_cell.Col = c Then Exit Do
        Set prev_cell = next_cell
    Loop
    
    ' Remove the cell from the column.
    Set prev_cell.NextInCol = next_cell.NextInCol

    ' If this is the only cell in the column,
    ' remove the column.
    If next_hdr.CellSentinel.NextInCol Is Nothing Then
        Set prev_hdr.NextHeader = next_hdr.NextHeader
    End If
End Sub

' ************************************************
' Set the indicated entry, creating a new cell
' if necessary.
' ************************************************
Property Let Value(r As Integer, c As Integer, new_value As Variant)
Dim prev_hdr As HeaderCell
Dim next_hdr As HeaderCell
Dim prev_cell As VerySparseArrayCell
Dim next_cell As VerySparseArrayCell
Dim new_cell As VerySparseArrayCell

    ' If new_value is NoValue, remove the entry.
    If new_value = NoValue Then
        RemoveEntry r, c
        Exit Property
    End If
    
    ' -------------------------
    ' Set the value in its row.
    ' -------------------------
    ' Find the row that should hold the entry.
    Set prev_hdr = RowSentinel
    Do
        Set next_hdr = prev_hdr.NextHeader
        If next_hdr Is Nothing Then Exit Do
        If next_hdr.RowOrCol > r Then
            Set next_hdr = Nothing
            Exit Do
        End If
        If next_hdr.RowOrCol = r Then Exit Do
        Set prev_hdr = next_hdr
    Loop
    
    ' If this row does not exist, create it.
    If next_hdr Is Nothing Then
        Set next_hdr = New HeaderCell
        Set next_hdr.CellSentinel = New VerySparseArrayCell
        Set next_hdr.NextHeader = prev_hdr.NextHeader
        Set prev_hdr.NextHeader = next_hdr
        next_hdr.RowOrCol = r
    End If

    ' Locate the cell in this row.
    Set prev_cell = next_hdr.CellSentinel
    Do
        Set next_cell = prev_cell.NextInRow
        If next_cell Is Nothing Then Exit Do
        If next_cell.Col > c Then
            Set next_cell = Nothing
            Exit Do
        End If
        If next_cell.Col = c Then Exit Do
        Set prev_cell = next_cell
    Loop
    
    ' If this cell does not exist, create it.
    If next_cell Is Nothing Then
        Set new_cell = New VerySparseArrayCell
        Set new_cell.NextInRow = prev_cell.NextInRow
        Set prev_cell.NextInRow = new_cell
        new_cell.Row = r
        new_cell.Col = c
        new_cell.Data = new_value
    Else
        ' Otherwise the cell already exists.
        ' Set its value and we're done.
        next_cell.Data = new_value
        Exit Property
    End If

    ' ----------------------------
    ' Set the value in its column.
    ' ----------------------------
    ' Find the column that should hold the entry.
    Set prev_hdr = ColSentinel
    Do
        Set next_hdr = prev_hdr.NextHeader
        If next_hdr Is Nothing Then Exit Do
        If next_hdr.RowOrCol > c Then
            Set next_hdr = Nothing
            Exit Do
        End If
        If next_hdr.RowOrCol = c Then Exit Do
        Set prev_hdr = next_hdr
    Loop
    
    ' If this column does not exist, create it.
    If next_hdr Is Nothing Then
        Set next_hdr = New HeaderCell
        Set next_hdr.CellSentinel = New VerySparseArrayCell
        Set next_hdr.NextHeader = prev_hdr.NextHeader
        Set prev_hdr.NextHeader = next_hdr
        next_hdr.RowOrCol = c
    End If

    ' Locate the cell in this column.
    Set prev_cell = next_hdr.CellSentinel
    Do
        Set next_cell = prev_cell.NextInRow
        If next_cell Is Nothing Then Exit Do
        If next_cell.Row > r Then
            Set next_cell = Nothing
            Exit Do
        End If
        If next_cell.Row = r Then Exit Do
        Set prev_cell = next_cell
    Loop

    ' If the cell does not exist, insert the
    ' previously created cell.
    If next_cell Is Nothing Then
        Set new_cell.NextInCol = prev_cell.NextInCol
        Set prev_cell.NextInCol = new_cell
    End If
End Property

' ************************************************
' Initialize NoValue.
' ************************************************
Private Sub Class_Initialize()
    NoValue = 0
End Sub

