Attribute VB_Name = "FStarStuff"
' ************************************************
' FStar.BAS
'
' Code for forward star N-ary tree example
' program.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

' Real nodes will run from 0 to NumNodes-1. Each
' of these will have an entry in FirstLink. The
' FirstLink array's sentinel will be at position
' NumNodes.
'
' Links will run from 0 to NumLinks - 1. Each of
' these will have an entry in ToNode. No sentinel
' is needed for the ToNode array.
Global FirstLink() As Integer
Global ToNode() As Integer
Global NumLinks As Integer
Global NumNodes As Integer

Global NodeCaption() As String
Global MaxLabel As Integer ' # of biggest label created

' ************************************************
' Add a new link between the indicated nodes.
' ************************************************
Sub AddLink(from_node As Integer, to_node As Integer)
Dim i As Integer

    ' Create room for the new link
    NumLinks = NumLinks + 1
    ReDim Preserve ToNode(0 To NumLinks - 1)

    ' Move the other links over to make room for the new one
    For i = NumLinks - 1 To FirstLink(from_node + 1) + 1 Step -1
        ToNode(i) = ToNode(i - 1)
    Next i

    ' Insert the new link
    ToNode(FirstLink(from_node + 1)) = to_node

    ' Update the FirstLink entries
    For i = from_node + 1 To NumNodes
        FirstLink(i) = FirstLink(i) + 1
    Next i
End Sub

' ************************************************
' Find the parent of this node and the link that
' connects it to its parent.
' ************************************************
Sub FindParent(node As Integer, parent As Integer, link As Integer)
    ' Find the link into this node
    For parent = 0 To NumNodes - 1
        For link = FirstLink(parent) To FirstLink(parent + 1) - 1
            If ToNode(link) = node Then Exit Sub
        Next link
    Next parent
    
    ' We did not find a parent
    parent = -1
    link = -1
End Sub

' ************************************************
' First recursively free this any descendants.
' Then free the node itself.
'
' Note that it is important that child nodes are
' always created after their parents in this
' program. That means when we recursively remove
' the children, the parent node does not move and
' its FirstLink entry does not move.
' ************************************************
Sub FreeNodeAndChildren(ByVal parent As Integer, ByVal link As Integer, ByVal node As Integer)
    ' Recursively remove the node's children.
    Do While FirstLink(node) < FirstLink(node + 1)
        FreeNodeAndChildren node, FirstLink(node), ToNode(FirstLink(node))
    Loop

    ' Remove the link.
    RemoveLink parent, link

    ' Remove the node itself.
    RemoveNode node
End Sub

' ************************************************
' Initialize sentinel at the end of the FirstLink
' array. Create the root node.
' ************************************************
Sub InitTree()
    ' Create the sentinel
    NumNodes = 1
    ReDim FirstLink(0 To NumNodes)
    ReDim NodeCaption(0 To NumNodes - 1)

    FirstLink(0) = 0    ' Root
    NodeCaption(0) = "0"
    
    FirstLink(1) = 0    ' Sentinel
End Sub

' ************************************************
' Create a new node at the end of the FirtsLink
' array.
' ************************************************
Function NewNode() As Integer
    ' Add an entry to the FirstLink array
    ReDim Preserve FirstLink(0 To NumNodes + 1)
    FirstLink(NumNodes + 1) = FirstLink(NumNodes)
    
    ' Load a new Text control
    Load FStarForm.NodeLabel(NumNodes)
    FStarForm.NodeLabel(NumNodes).BackColor = FStarForm.BackColor
    FStarForm.NodeLabel(NumNodes).ForeColor = FStarForm.ForeColor
    FStarForm.NodeLabel(NumNodes).Visible = True
    
    ' Create a new label value
    MaxLabel = MaxLabel + 1
    ReDim Preserve NodeCaption(0 To NumNodes)
    NodeCaption(NumNodes) = Format$(MaxLabel)
    
    NewNode = NumNodes
    NumNodes = NumNodes + 1
End Function

' ************************************************
' Remove the indicated link.
' ************************************************
Sub RemoveLink(node As Integer, link As Integer)
Dim i As Integer

    ' Update FirstLink entries.
    For i = node + 1 To NumNodes
        FirstLink(i) = FirstLink(i) - 1
    Next i

    ' Shift ToNode array to fill in the link's
    ' spot.
    For i = link + 1 To NumLinks - 1
        ToNode(i - 1) = ToNode(i)
    Next i

    ' Remove the extra position in ToNode.
    NumLinks = NumLinks - 1
    If NumLinks > 0 Then ReDim Preserve ToNode(0 To NumLinks - 1)
End Sub

' ************************************************
' Remove a node from the tree. This routine
' assumes that the node has no links and no
' parent.
' ************************************************
Sub RemoveNode(node As Integer)
Dim i As Integer

    ' Slide FirstLink entries over to fill in
    ' the vacated spot.
    For i = node + 1 To NumNodes
        FirstLink(i - 1) = FirstLink(i)
    Next i

    ' Slide NodeCaption entries over.
    For i = node + 1 To NumNodes - 1
        NodeCaption(i - 1) = NodeCaption(i)
    Next i

    ' Update ToNode entries pointing to nodes
    ' after this.
    For i = 0 To NumLinks - 1
        If ToNode(i) >= node Then ToNode(i) = ToNode(i) - 1
    Next i
    
    ' Remove the extra FirstLink entry.
    NumNodes = NumNodes - 1
    ReDim Preserve FirstLink(0 To NumNodes)
    
    ReDim Preserve NodeCaption(0 To NumNodes - 1)
    Unload FStarForm.NodeLabel(NumNodes)
End Sub

