Attribute VB_Name = "QtreeStuff"
' ************************************************
' Qtree2.BAS
'
' Code for pointer faking quadtree program.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

Global Const END_OF_LIST = 32767

Global Const MAX_PER_NODE = 100 ' Max items per node.

' Quadtree nodes.
'
' If this is a leaf node, SWChild = END_OF_LIST and
' NWChild stores the index of the first item in the
' item linked list.
Type QtreeNode
    NWchild As Long
    NEchild As Long
    SEchild As Long
    SWchild As Long
    NumItems As Long
End Type

' Linked list of items within each leaf node.
Type QtreeItem
    X As Single
    Y As Single
    NextItem As Long
End Type

' Quadtree variables.
Dim NumInQtree As Long       ' # nodes allocated.
Dim NextQtree As Long        ' Index of next unused node.
Global QtreeNodes() As QtreeNode ' Quadtree nodes.

' Item linked list variables.
Global Const MIN_ITEM = -30000
Dim MaxItem As Long          ' Largest item index allocated.
Global NextItem As Long      ' Index of next unused item.
Global Items() As QtreeItem     ' Array of items.

' Drawing area bounds.
Global Gxmin As Single
Global Gxmax As Single
Global Gymin As Single
Global Gymax As Single

' ************************************************
' Check leaves near the indicated leaf to see if
' they have items closer to the given item than
' the best found so far.
' ************************************************
Sub CheckNearbyLeaves(node As Long, exclude As Long, X As Single, Y As Single, best_item As Long, best_dist As Single, comparisons As Long, xmin As Single, xmax As Single, ymin As Single, ymax As Single)
Dim xmid As Single
Dim ymid As Single
Dim new_dist As Single
Dim new_item As Long

    ' If this is the leaf we are to exclude, stop now.
    If node = exclude Then Exit Sub

    ' If this is a leaf node, check it out.
    If QtreeNodes(node).SWchild = END_OF_LIST Then
        NearPointInLeaf node, X, Y, new_item, new_dist, comparisons
        If best_dist > new_dist Then
            best_dist = new_dist
            best_item = new_item
        End If
        Exit Sub
    End If
    
    ' See which children fall within best_dist of the point.
    xmid = (xmax + xmin) / 2
    ymid = (ymax + ymin) / 2
    If X - Sqr(best_dist) <= xmid Then
        ' The West children are eligible.
        If Y - Sqr(best_dist) <= ymid Then
            ' Check the NorthWest child.
            CheckNearbyLeaves QtreeNodes(node).NWchild, exclude, X, Y, best_item, best_dist, comparisons, xmin, xmid, ymin, ymid
        End If
        If Y + Sqr(best_dist) > ymid Then
            ' Check the SouthWest child.
            CheckNearbyLeaves QtreeNodes(node).SWchild, exclude, X, Y, best_item, best_dist, comparisons, xmin, xmid, ymid, ymax
        End If
    End If
    If X + Sqr(best_dist) > xmid Then
        ' The East children are eligible.
        If Y - Sqr(best_dist) <= ymid Then
            ' Check the NorthEast child.
            CheckNearbyLeaves QtreeNodes(node).NEchild, exclude, X, Y, best_item, best_dist, comparisons, xmid, xmax, ymin, ymid
        End If
        If Y + Sqr(best_dist) > ymid Then
            ' Check the SouthEast child.
            CheckNearbyLeaves QtreeNodes(node).SEchild, exclude, X, Y, best_item, best_dist, comparisons, xmid, xmax, ymid, ymax
        End If
    End If
End Sub

' ************************************************
' Divide this node if necessary.
'
' The Northwest child holds the item list.
' ************************************************
Sub DivideNode(ByVal node As Long, xmin As Single, xmax As Single, ymin As Single, ymax As Single)
Dim xmid As Single
Dim ymid As Single
Dim item As Long
Dim nxt As Long
Dim child As Long

    If QtreeNodes(node).NumItems <= MAX_PER_NODE Then Exit Sub

    ' Save the index of the first item.
    nxt = QtreeNodes(node).NWchild

    ' Create the children.
    QtreeNodes(node).NWchild = NewQtreeNode()
    QtreeNodes(node).NEchild = NewQtreeNode()
    QtreeNodes(node).SEchild = NewQtreeNode()
    QtreeNodes(node).SWchild = NewQtreeNode()

    ' Divide the children up.
    xmid = (xmin + xmax) / 2
    ymid = (ymin + ymax) / 2
    Do While nxt <> END_OF_LIST
        item = nxt
        nxt = Items(item).NextItem
        ' See in which child the item belongs.
        If Items(item).X <= xmid Then
            If Items(item).Y <= ymid Then
                ' NW child.
                child = QtreeNodes(node).NWchild
            Else
                ' SW child.
                child = QtreeNodes(node).SWchild
            End If
        Else
            If Items(item).Y <= ymid Then
                ' NE child.
                child = QtreeNodes(node).NEchild
            Else
                ' SE child.
                child = QtreeNodes(node).SEchild
            End If
        End If

        ' Place the item in its new node.
        Items(item).NextItem = QtreeNodes(child).NWchild
        QtreeNodes(child).NWchild = item
        QtreeNodes(child).NumItems = QtreeNodes(child).NumItems + 1
    Loop

    ' Recursively divide the children if necessary.
    DivideNode QtreeNodes(node).NWchild, xmin, xmid, ymin, ymid
    DivideNode QtreeNodes(node).NEchild, xmid, xmax, ymin, ymid
    DivideNode QtreeNodes(node).SWchild, xmin, xmid, ymid, ymax
    DivideNode QtreeNodes(node).SEchild, xmid, xmax, ymid, ymax
End Sub

' ************************************************
' Use the quadtree to locate the point.
' ************************************************
Function FindPoint(X As Single, Y As Single, comparisons As Long) As Long
Dim leaf As Long
Dim item As Long
Dim dist As Single

    ' See what leaf the point is in.
    leaf = LocateLeaf(X, Y)

    ' Find the closest point within the leaf.
    NearPointInLeaf leaf, X, Y, item, dist, comparisons

    ' Check nearby leaves.
    CheckNearbyLeaves 1, leaf, X, Y, item, dist, comparisons, Gxmin, Gxmax, Gymin, Gymax

    FindPoint = item
End Function

' ************************************************
' Find the quadtree leaf node that includes this
' one.
' ************************************************
Function LocateLeaf(X As Single, Y As Single) As Long
Dim xmin As Single
Dim xmid As Single
Dim xmax As Single
Dim ymin As Single
Dim ymid As Single
Dim ymax As Single
Dim node As Long
    
    xmin = Gxmin
    xmax = Gxmax
    ymin = Gymin
    ymax = Gymax
    node = 1    ' Start at the root.
    ' Loop until we hit a leaf.
    Do While QtreeNodes(node).SWchild <> END_OF_LIST
        xmid = (xmax + xmin) / 2
        ymid = (ymax + ymin) / 2
        If X <= xmid Then
            If Y <= ymid Then
                node = QtreeNodes(node).NWchild
                ymax = ymid
            Else
                node = QtreeNodes(node).SWchild
                ymin = ymid
            End If
            xmax = xmid
        Else
            If Y <= ymid Then
                node = QtreeNodes(node).NEchild
                ymax = ymid
            Else
                node = QtreeNodes(node).SEchild
                ymin = ymid
            End If
            xmin = xmid
        End If
    Loop
    LocateLeaf = node
End Function

' ************************************************
' Find the closest item to the indicated point
' within this leaf node.
' ************************************************
Sub NearPointInLeaf(leaf As Long, X As Single, Y As Single, best_item As Long, best_dist As Single, comparisons As Long)
Dim item As Long
Dim Dx As Single
Dim Dy As Single
Dim new_dist As Single

    best_item = QtreeNodes(leaf).NWchild
    
    ' If there are no items in the leaf, stop now.
    If best_item = END_OF_LIST Then
        best_dist = 10000000
        Exit Sub
    End If

    comparisons = comparisons + 1
    Dx = Items(best_item).X - X
    Dy = Items(best_item).Y - Y
    best_dist = Dx * Dx + Dy * Dy
    item = Items(best_item).NextItem
    Do While item <> END_OF_LIST
        comparisons = comparisons + 1
        Dx = Items(item).X - X
        Dy = Items(item).Y - Y
        new_dist = Dx * Dx + Dy * Dy
        If best_dist > new_dist Then
            best_dist = new_dist
            best_item = item
        End If
        item = Items(item).NextItem
    Loop
End Sub

' ************************************************
' Allocate a new data item.
' ************************************************
Function NewItem() As Long
    If NextItem > MaxItem Then
         MaxItem = MaxItem + 100
         ReDim Preserve Items(MIN_ITEM To MaxItem)
    End If
    
    ' Initialize the children.
    Items(NextItem).NextItem = END_OF_LIST

    NewItem = NextItem
    NextItem = NextItem + 1
End Function

' ************************************************
' Allocate a new quadtree node.
' ************************************************
Function NewQtreeNode() As Long
    If NextQtree > NumInQtree Then
         NumInQtree = NextQtree + 10
         ReDim Preserve QtreeNodes(1 To NumInQtree)
    End If
    
    ' Initialize the children.
    QtreeNodes(NextQtree).NWchild = END_OF_LIST
    QtreeNodes(NextQtree).NEchild = END_OF_LIST
    QtreeNodes(NextQtree).SEchild = END_OF_LIST
    QtreeNodes(NextQtree).SWchild = END_OF_LIST
    QtreeNodes(NextQtree).NumItems = 0
    
    NewQtreeNode = NextQtree
    NextQtree = NextQtree + 1
End Function

' ************************************************
' Initialize the quadtree so it is empty.
' ************************************************
Sub QtreeInitialize()
Dim Root As Long

    ' Allocate the quadtree root.
    NextQtree = 1
    Root = NewQtreeNode()

    ' Force a resize when we get the first item.
    NextItem = MIN_ITEM
    MaxItem = MIN_ITEM - 1
End Sub

