Attribute VB_Name = "BplusStuff"
' ************************************************
' Bplus.BAS
'
' Source code for B+Tree example program.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

' Indicates no child.
Global Const NO_CHILD = 0

' Record to hold customer info in Custs.dat.
Type CustRecord
    LastName As String * 20
    FirstName As String * 20
    Address As String * 40
    City As String * 20
    State As String * 2
    Zip As String * 10
    Phone As String * 12
    NextGarbage As Long
End Type

' Size of the customer record.
Global Const CUST_SIZE = _
    20 + 20 + 40 + 20 + 2 + 10 + 12 + 4

' We use FirstName,LastName as the bucket keys.
' These, plus the comma, have a size of 41 bytes.
' Adding in the Long child pointers we can fit 22
' keys into 1,024 bytes blocks. Therefore we will
' make the tree have order 11 and we will put 22
' keys in each bucket.
Const KEY_SIZE = 41
Const ORDER = 11
'Const ORDER = 2 ' Make ORDER = 2 for easier testing.
Global Const KEYS_PER_NODE = 2 * ORDER

Type Bucket
    NumKeys As Integer
    Key(1 To KEYS_PER_NODE) As String * KEY_SIZE
    Child(0 To KEYS_PER_NODE) As Long
End Type
Global Const BUCKET_SIZE = 2 + _
    KEYS_PER_NODE * KEY_SIZE + _
    (KEYS_PER_NODE + 1) * 4

' This record contains header information plus
' enough padding to make it the same size as a
' bucket.
Global Const HEADER_PADDING = _
    BUCKET_SIZE - (7 * 4 + 2)
Type HeaderRecord
    NumBuckets As Long
    NumRecords As Long
    Root As Long
    NextTreeRecord As Long
    NextCustRecord As Long
    FirstTreeGarbage As Long
    FirstCustGarbage As Long
    Height As Integer
    Padding As String * HEADER_PADDING
End Type

Global Header As HeaderRecord       ' Cached header info.
Global HeaderModified As Integer    ' Has it been changed?
Global CachedBuckets() As Bucket    ' Cached buckets.
Global CachedIndexes() As Long      ' Indexes of cached buckets.

Global DatFile As Long              ' Custs.dat file handle.
Global IdxFile As Long              ' Custs.idx file handle.

Global CurrentRecord As CustRecord  ' Record currently displayed.
Global CurrentIndex As Long         ' Index of CurrentRecord.

Global DiskAccesses As Integer      ' # trips to disk.
' ************************************************

' ************************************************
' Cache this node if it is not already cached.
' Return True if there is an error, False
' otherwise.
' ************************************************
Private Function CacheBucket(depth As Integer, node As Long) As Integer
    CacheBucket = False
    
    ' If the bucket is already cached, do nothing.
    If CachedIndexes(depth) = node Then Exit Function
    
    ' Read the bucket.
    GetBucket CachedBuckets(depth), node
    CachedIndexes(depth) = node
End Function

' ************************************************
' Close files Custs.dat and Custs.idx.
' ************************************************
Public Sub CloseFiles()
    On Error GoTo CloseErr
    Close IdxFile
    Close DatFile
    Exit Sub

CloseErr:
    Beep
    MsgBox "Error closing B+Tree files.", vbExclamation, "Severe Error"
    Resume Next
End Sub

' ************************************************
' Delete an item from the tree.
' ************************************************
Public Sub DeleteItem(lname As String, fname As String)
Dim txt As String
Dim too_small As Integer
Dim old_root As Long
Dim status As Integer
Dim up_value As String

    txt = lname & "," & fname

    ' Delete the item.
    DeleteItemFromNode 1, txt, lname, fname, too_small, up_value
    
    ' If the root has shrunk, see if we need to remove
    ' a level from the tree.
    If CachedBuckets(1).NumKeys < 1 Then
        ' Remember the old root so we can free it.
        old_root = Header.Root

        ' Update the header.
        Header.Height = Header.Height - 1
        Header.Root = CachedBuckets(1).Child(0)
        
        ' Free the old root.
        FreeBucket old_root
        
        ' See if the tree is empty.
        If Header.Height > 0 Then
            ' If the tree is not empty, cache the new root.
            status = CacheBucket(1, Header.Root)
        Else
            ' Otherwise reset everything.
            Header.Root = NO_CHILD   ' No root node.
            Header.NextTreeRecord = 1   ' Next bucket available.
            Header.NextCustRecord = 0   ' Next cust available.
            Header.FirstTreeGarbage = NO_CHILD   ' No garbage yet.
            Header.FirstCustGarbage = NO_CHILD   ' No garbage yet.
        End If

        ' Save the updated header information.
        WriteHeader
    End If
End Sub

' ************************************************
' Recursively remove the item from somewhere down
' this node's branches.
' ************************************************
Private Sub DeleteItemFromNode(depth As Integer, txt As String, lname As String, fname As String, too_small As Integer, up_value As String)
Dim branch As Integer
Dim node As Long
Dim i As Integer
Dim cust As CustRecord

    ' See down which branch we should search.
    For branch = 1 To CachedBuckets(depth).NumKeys
        If CachedBuckets(depth).Key(branch) > txt Then Exit For
    Next branch
    
    ' If we are at a leaf, this had better be it.
    If depth >= Header.Height Then
        ' Make depth point to the customer record which
        ' should match the target.
        branch = branch - 1

        ' Get the customer record and see if it matches.
        node = CachedBuckets(depth).Child(branch)
        If node <> NO_CHILD Then
            GetCust cust, node
            If cust.LastName <> lname Or cust.FirstName <> fname Then node = NO_CHILD
        End If
        If node = NO_CHILD Then
            Beep
            MsgBox "Record """ & Trim$(lname) & "," & Trim$(fname) & """ is not in the database."
            too_small = False
            Exit Sub
        End If
    
        ' At this point we have verified that this is
        ' the target. Since this is a B+Tree, we know we
        ' are at a leaf. We simply remove the leaf entry
        ' and return up the tree.
        
        ' If this is the leftmost entry in the leaf, we
        ' must revise the ancestor that holds the target
        ' value as a key.
        If branch = 0 Then
            up_value = CachedBuckets(depth).Key(1)
        Else
            up_value = ""
        End If
        
        ' Remove the entry from the leaf node.
        CachedBuckets(depth).NumKeys = CachedBuckets(depth).NumKeys - 1
        too_small = (CachedBuckets(depth).NumKeys < ORDER)
        
        If branch <= CachedBuckets(depth).NumKeys Then
            CachedBuckets(depth).Child(branch) = CachedBuckets(depth).Child(branch + 1)
            If branch > 0 Then CachedBuckets(depth).Key(branch) = CachedBuckets(depth).Key(branch + 1)
        End If

        For i = branch + 1 To CachedBuckets(depth).NumKeys
            CachedBuckets(depth).Key(i) = CachedBuckets(depth).Key(i + 1)
            CachedBuckets(depth).Child(i) = CachedBuckets(depth).Child(i + 1)
        Next i
        
        ' Save the updated leaf node to disk.
        WriteBucket CachedBuckets(depth), CachedIndexes(depth)
        
        ' Free the customer record.
        FreeCust node

        ' We are done performing the actual deletion.
        Exit Sub
    End If

    ' Continue down the tree looking for the item.
    
    ' Cache the next bucket in the search.
    node = CachedBuckets(depth).Child(branch - 1)
    If CacheBucket(depth + 1, node) Then
        too_small = False
        Exit Sub
    End If

    ' Delete the item from that bucket.
    DeleteItemFromNode depth + 1, txt, lname, fname, too_small, up_value

    ' If this node holds the exact value, update it with
    ' up_value.
    If branch > 1 And up_value <> "" Then
        If CachedBuckets(depth).Key(branch - 1) = txt Then
            CachedBuckets(depth).Key(branch - 1) = up_value
            up_value = ""
        End If
    End If

    ' If the child node is now too small, rearrange it.
    If too_small Then TooSmall CachedBuckets(depth), CachedIndexes(depth), CachedBuckets(depth + 1), CachedIndexes(depth + 1), branch - 1, too_small
End Sub

' ************************************************
' Return a bucket to the garbage bucket list. The
' calling routine must save the header information
' to disk.
' ************************************************
Private Sub FreeBucket(node As Long)
Dim buck As Bucket

    buck.Child(0) = Header.FirstTreeGarbage
    WriteBucket buck, node
    Header.FirstTreeGarbage = node
    Header.NumBuckets = Header.NumBuckets - 1
End Sub

' ************************************************
' Return a customer record to the garbage customer
' record list. The calling routine must save the
' header information to disk.
' ************************************************
Private Sub FreeCust(node As Long)
Dim cust  As CustRecord

    cust.NextGarbage = Header.FirstCustGarbage
    WriteCust cust, node
    Header.FirstCustGarbage = node
    Header.NumRecords = Header.NumRecords - 1
End Sub

' ************************************************
' Get this bucket from the file. Return True if
' there is a problem, False otherwise.
' ************************************************
Public Sub GetBucket(rec As Bucket, recnum As Long)
Dim buf As Bucket

    DiskAccesses = DiskAccesses + 1
    Get #IdxFile, recnum, buf
    rec = buf
End Sub

' ************************************************
' Get this customer record from the file. Return
' True if there is a problem, False otherwise.
' ************************************************
Public Sub GetCust(rec As CustRecord, recnum As Long)
Dim buf As CustRecord

    DiskAccesses = DiskAccesses + 1
    Get #DatFile, recnum, buf
    rec = buf
End Sub

' ************************************************
' Get the tree header from the first record in the
' B+Tree file. Return True if there is a problem,
' False otherwise.
' ************************************************
Private Sub GetHeader()
    DiskAccesses = DiskAccesses + 1
    Get #IdxFile, 1, Header
End Sub

' ************************************************
' Insert an item in a node known to have an empty
' position.
' ************************************************
Private Sub InsertInNodeWithRoom(depth As Integer, node As Long, spot As Integer, new_key As String, new_child As Long)
Dim i As Integer

    ' Make room for the new entry.
    CachedBuckets(depth).NumKeys = CachedBuckets(depth).NumKeys + 1
    For i = CachedBuckets(depth).NumKeys To spot + 1 Step -1
        CachedBuckets(depth).Key(i) = CachedBuckets(depth).Key(i - 1)
        CachedBuckets(depth).Child(i) = CachedBuckets(depth).Child(i - 1)
    Next i
    
    ' Insert the new entry.
    CachedBuckets(depth).Key(spot) = new_key
    CachedBuckets(depth).Child(spot) = new_child

    ' Update the record to disk.
    WriteBucket CachedBuckets(depth), CachedIndexes(depth)
End Sub

' ************************************************
' Insert an item into the tree.
' ************************************************
Public Sub InsertItem(txt As String)
Dim up_node As Long
Dim up_key As String
Dim had_split As Integer
Dim old_root As Long
Dim i As Integer

    HeaderModified = False

    ' Add the new item.
    InsertItemInNode 1, Header.Root, txt, up_node, up_key, had_split
    
    ' If the root node split, create a new root that
    ' contains the item that was sent up to the parent.
    If had_split Then
        ' Make room for more cached buckets.
        ReDim Preserve CachedBuckets(1 To Header.Height + 1)
        ReDim Preserve CachedIndexes(1 To Header.Height + 1)

        ' Create the new root.
        old_root = Header.Root
        CachedIndexes(1) = NewBucket(CachedBuckets(1))
        CachedBuckets(1).Key(1) = up_key
        CachedBuckets(1).Child(0) = old_root
        CachedBuckets(1).Child(1) = up_node
        CachedBuckets(1).NumKeys = 1
        WriteBucket CachedBuckets(1), CachedIndexes(1)
        
        ' Prepare to update the header information.
        Header.Root = CachedIndexes(1)
        Header.Height = Header.Height + 1
    End If
    ' In either case the header was modified. At least
    ' FirstCustGarbage has changed.
    WriteHeader
End Sub

' ************************************************
' Recursively insert the item below this node. If
' there is a split, add the item passed up to this
' node. If this node is then full, split it.
' ************************************************
Private Sub InsertItemInNode(depth As Integer, node As Long, new_key As String, up_node As Long, up_key As String, had_split As Integer)
Dim branch As Integer
Dim i As Integer
Dim return_node As Long

    If depth > Header.Height Then
        ' We're at the bottom of the tree. Create the
        ' new customer record and pass its record number
        ' back up for insertion in the node above.
        CurrentIndex = NewCust(CurrentRecord)
        up_node = CurrentIndex
        up_key = new_key
        had_split = True
        Exit Sub
    End If

    ' Cache the bucket.
    If CacheBucket(depth, node) Then Exit Sub

    ' See which branch we should investigate.
    For branch = 0 To CachedBuckets(depth).NumKeys - 1
        If CachedBuckets(depth).Key(branch + 1) > new_key Then Exit For
    Next branch
    
    ' Branch.
    InsertItemInNode depth + 1, CachedBuckets(depth).Child(branch), new_key, up_node, up_key, had_split

    ' If we had a split, deal with it.
    If had_split Then
        If CachedBuckets(depth).NumKeys < KEYS_PER_NODE Then
            ' If there's room, just insert the item.
            InsertInNodeWithRoom depth, node, branch + 1, up_key, up_node
            had_split = False
        Else
            ' The node is full. Split it.
            SplitBucket depth, node, branch + 1, up_key, up_node
        End If
    End If
End Sub

' ************************************************
' Allocate a new B+Tree bucket. Return the record
' number for the bucket.
'
' The calling routine must save the updated header
' information and the new bucket to disk.
' ************************************************
Private Function NewBucket(buck As Bucket) As Long
Dim recnum As Long
Dim i As Integer

    If Header.FirstTreeGarbage = NO_CHILD Then
        ' If the garbage list is empty, take a new bucket
        ' from the end of the file.
        recnum = Header.NextTreeRecord
        Header.NextTreeRecord = Header.NextTreeRecord + 1
    Else
        ' Reuse the top garbage item.
        recnum = Header.FirstTreeGarbage
        GetBucket buck, recnum
        Header.FirstTreeGarbage = buck.Child(0)
    End If
    Header.NumBuckets = Header.NumBuckets + 1
    
    ' Blank the new bucket.
    buck.NumKeys = 0
    
    NewBucket = recnum
End Function

' ************************************************
' Allocate a new customer record and fill in the
' fields from cust. Return the record number of
' the new record.
'
' The calling routine must save the updated header
' information to disk.
' ************************************************
Private Function NewCust(cust As CustRecord) As Long
Dim recnum As Long
Dim i As Integer
Dim garbage As CustRecord

    If Header.FirstCustGarbage = NO_CHILD Then
        ' If the garbage list is empty, take a new
        ' bucket from the end of the file.
        recnum = Header.NextCustRecord
        Header.NextCustRecord = Header.NextCustRecord + 1
    Else
        ' Reuse the top garbage item.
        recnum = Header.FirstCustGarbage
        GetCust garbage, recnum
        Header.FirstCustGarbage = garbage.NextGarbage
    End If
    Header.NumRecords = Header.NumRecords + 1

    ' Update the new record.
    WriteCust cust, recnum
    
    NewCust = recnum
End Function

' ************************************************
' Open files Custs.dat and Custs.idx.
'
' If the files do not exist, create them and
' initialize the header information.
'
' Cache the root bucket.
' ************************************************
Public Sub OpenFiles()
Dim filename As String
Dim file_path As String

    file_path = InputBox$("Data directory", _
        "Data Directory", App.Path)
    If Right$(file_path, 1) <> "\" Then _
        file_path = file_path & "\"
    
    CurrentIndex = NO_CHILD   ' No record displayed.

    ' Open Custs.dat.
    On Error GoTo OpenErr
    filename = file_path & "Custs.dat"
    DatFile = FreeFile
    Open filename For Random As DatFile Len = CUST_SIZE

    ' Open Custs.idx.
    filename = file_path & "Custs.idx"
    IdxFile = FreeFile
    Open filename For Random As IdxFile Len = BUCKET_SIZE
    
    ' See if the file has just been created.
    If LOF(IdxFile) = 0 Then
        ' Initialize the header information.
        Header.Height = 0           ' Tree is empty.
        Header.Root = NO_CHILD   ' No root node.
        Header.NextTreeRecord = 2   ' Next bucket available.
        Header.NextCustRecord = 1   ' Next cust available.
        Header.FirstTreeGarbage = NO_CHILD   ' No garbage yet.
        Header.FirstCustGarbage = NO_CHILD   ' No garbage yet.

        ' Write the header to disk.
        WriteHeader
    Else
        ' Read the header information.
        GetHeader
    End If

    ' Allocate the bucket cache.
    If Header.Height > 0 Then
        ReDim CachedBuckets(1 To Header.Height)
        ReDim CachedIndexes(1 To Header.Height)
        
        ' Cache the root bucket.
        GetBucket CachedBuckets(1), Header.Root
        CachedIndexes(1) = Header.Root
    End If
    
    Exit Sub

OpenErr:
    Beep
    MsgBox "Error opening B+tree files. Exiting.", vbExclamation, "Fatal Error"
    End
End Sub

' ************************************************
' Split the node plus the up_key item into two
' nodes. Variable spot indicates where in the node
' the up_key item should be placed.
'
' Return the new node through parameter
' return_node. Return the key to be passed back up
' through parameter up_key.
' ************************************************
Private Sub SplitBucket(depth As Integer, node As Long, spot As Integer, up_key As String, up_node As Long)
Dim i As Integer
Dim return_node As Long
Dim return_bucket As Bucket
Dim return_key As String
Dim right_child_0 As Long

    ' Create a new bucket to the right.
    return_node = NewBucket(return_bucket)

    If spot <= ORDER + 1 Then
        ' The new item should be in the left node or
        ' should be passed up.
        If spot = ORDER + 1 Then
            return_key = up_key
            right_child_0 = up_node
        Else
            ' Save the ORDER-th entry.
            return_key = CachedBuckets(depth).Key(ORDER)
            right_child_0 = CachedBuckets(depth).Child(ORDER)

            ' Insert the new item into the left node.
            For i = ORDER To spot + 1 Step -1
                CachedBuckets(depth).Key(i) = CachedBuckets(depth).Key(i - 1)
                CachedBuckets(depth).Child(i) = CachedBuckets(depth).Child(i - 1)
            Next i
            CachedBuckets(depth).Key(spot) = up_key
            CachedBuckets(depth).Child(spot) = up_node
        End If
        ' Move items into the right node.
        For i = 1 To ORDER
            return_bucket.Key(i) = CachedBuckets(depth).Key(i + ORDER)
            return_bucket.Child(i) = CachedBuckets(depth).Child(i + ORDER)
        Next i
    Else
        ' The new item should be in the right node.
        spot = spot - ORDER - 1
        
        return_key = CachedBuckets(depth).Key(ORDER + 1)
        right_child_0 = CachedBuckets(depth).Child(ORDER + 1)

        ' Move the first set of items into the right node.
        For i = 1 To spot - 1
            return_bucket.Key(i) = CachedBuckets(depth).Key(i + ORDER + 1)
            return_bucket.Child(i) = CachedBuckets(depth).Child(i + ORDER + 1)
        Next i
        
        ' Put the up_key/up_node in the right node.
        return_bucket.Key(spot) = up_key
        return_bucket.Child(spot) = up_node
        
        ' Put the other items in the right node.
        For i = spot + 1 To ORDER
            return_bucket.Key(i) = CachedBuckets(depth).Key(i + ORDER)
            return_bucket.Child(i) = CachedBuckets(depth).Child(i + ORDER)
        Next i
    End If
    
    CachedBuckets(depth).NumKeys = ORDER
    return_bucket.NumKeys = ORDER

    return_bucket.Child(0) = right_child_0
    
    up_node = return_node
    up_key = return_key

    ' Update the split buckets.
    WriteBucket return_bucket, return_node
    WriteBucket CachedBuckets(depth), CachedIndexes(depth)
End Sub

' ************************************************
' The node "small" is too small. It is located in
' position "spot" in the parent node "parent".
'
' If node has a right sibling, try to redistribute
' items between node and the right sibling. If the
' right sibling has too few items (if it has
' exactly ORDER items), merge node and the right
' sibling.
'
' If the node does not have a right sibling,
' perform the steps above for the left sibling.
'
' If two nodes are merged, see if the parent is
' now too small.
'
' The calling routine must save the header
' information to disk. It will be updated if a
' bucket is freed.
' ************************************************
Private Sub TooSmall(parent As Bucket, parent_index As Long, small As Bucket, small_index As Long, spot As Integer, too_small As Integer)
Dim node As Long
Dim num_in_parent As Integer
Dim num_in_sibling As Integer
Dim sib_index As Long
Dim sibling As Bucket
Dim num_to_move As Integer
Dim i As Integer

    num_in_parent = parent.NumKeys

    ' If there is a right sibling...
    If spot < num_in_parent Then
        ' Look at the right sibling.
        spot = spot + 1     ' Now points to the sibling.
        sib_index = parent.Child(spot)  ' Sibling's index.
        GetBucket sibling, sib_index
        num_in_sibling = sibling.NumKeys
        num_to_move = (num_in_sibling - ORDER + 1) \ 2

        ' Put item from parent at end of items in the
        ' small node. Right now it has ORDER - 1 keys.
        small.Key(ORDER) = parent.Key(spot)
        small.Child(ORDER) = sibling.Child(0)

        If num_to_move > 0 Then
            ' If we have enough items to redistribute,
            ' do so. Copy items from sibling to small.
            For i = 1 To num_to_move - 1
                small.Key(i + ORDER) = sibling.Key(i)
                small.Child(i + ORDER) = sibling.Child(i)
            Next i

            ' Prepare the parent.
            parent.Key(spot) = sibling.Key(num_to_move)
            parent.Child(spot) = sib_index

            ' Fill the hole in the sibling.
            sibling.Child(0) = sibling.Child(num_to_move)
            num_in_sibling = num_in_sibling - num_to_move
            For i = 1 To num_in_sibling
                sibling.Key(i) = sibling.Key(i + num_to_move)
                sibling.Child(i) = sibling.Child(i + num_to_move)
            Next i

            ' Update the key counts.
            sibling.NumKeys = num_in_sibling
            small.NumKeys = ORDER - 1 + num_to_move
            too_small = False

            ' Save the updated buckets to disk.
            WriteBucket parent, parent_index
            WriteBucket small, small_index
            WriteBucket sibling, sib_index
        Else ' End if enough items to redistribute...
            ' If there are not enough items to
            ' redistribute, merge the nodes.

            ' Move entries from sibling to small. It must
            ' contain ORDER items.
            For i = 1 To ORDER
                small.Key(i + ORDER) = sibling.Key(i)
                small.Child(i + ORDER) = sibling.Child(i)
            Next i

            ' Fill in the hole in the parent.
            For i = spot To num_in_parent - 1
                parent.Key(i) = parent.Key(i + 1)
                parent.Child(i) = parent.Child(i + 1)
            Next i

            ' Update the key counts.
            small.NumKeys = KEYS_PER_NODE
            parent.NumKeys = num_in_parent - 1

            ' Free the sibling node.
            FreeBucket sib_index
            too_small = (parent.NumKeys < ORDER)

            ' Save the updated buckets to disk.
            WriteBucket parent, parent_index
            WriteBucket small, small_index
        End If  ' End if (can redistribute) ... else (merge)
    Else ' End looking at right sibling
        ' There is no right sibling. Look at the left
        ' sibling.
        sib_index = parent.Child(spot - 1)
        GetBucket sibling, sib_index
        num_in_sibling = sibling.NumKeys + 1
        num_to_move = (num_in_sibling - ORDER) \ 2
        If num_to_move > 0 Then
            ' If we have enough items to redistribute,
            ' do so.

            ' Make room for the new items in the node.
            For i = ORDER - 1 To 1 Step -1
                small.Key(i + num_to_move) = small.Key(i)
                small.Child(i + num_to_move) = small.Child(i)
            Next i

            ' Move an item from the parent into the node.
            small.Key(num_to_move) = parent.Key(spot)
            small.Child(num_to_move) = small.Child(0)

            ' Move items from the sibling into the node.
            num_in_sibling = num_in_sibling - num_to_move
            For i = num_to_move - 1 To 1 Step -1
                small.Key(i) = sibling.Key(i + num_in_sibling)
                small.Child(i) = sibling.Child(i + num_in_sibling)
            Next i
            small.Child(0) = sibling.Child(num_in_sibling)

            ' Revise the sibling's entry in the parent.
            parent.Key(spot) = sibling.Key(num_in_sibling)
            
            ' Update the key counts.
            sibling.NumKeys = num_in_sibling - 1
            small.NumKeys = ORDER - 1 + num_to_move
            too_small = False
            
            ' Save the updated buckets to disk.
            WriteBucket parent, parent_index
            WriteBucket small, small_index
            WriteBucket sibling, sib_index
        Else ' End if enough items to redistribute
            ' Not enough items to redistribute.
            ' Merge the nodes.

            ' Put the node's parent entry in the sibling.
            sibling.Key(num_in_sibling) = parent.Key(spot)
            sibling.Child(num_in_sibling) = small.Child(0)

            ' Move items from the node into the sibling.
            For i = 1 To ORDER - 1
                sibling.Key(i + num_in_sibling) = small.Key(i)
                sibling.Child(i + num_in_sibling) = small.Child(i)
            Next i

            ' Update the key counts.
            sibling.NumKeys = KEYS_PER_NODE
            parent.NumKeys = num_in_parent - 1

            ' Free the node.
            FreeBucket small_index
            too_small = (parent.NumKeys < ORDER)

            ' Save the updated buckets to disk.
            WriteBucket parent, parent_index
            WriteBucket sibling, sib_index
        End If  ' End if (can redistribute) ... else (merge)
    End If ' End if (right sibling) ... else (left sibling)
End Sub

' ************************************************
' Create a string representing the tree's
' structure.
' ************************************************
Public Function TreeStructure(spaces As Integer, depth As Integer, lines As Integer, show_leaves As Integer) As String
Const MAX_LINES = 500

Dim txt As String
Dim i As Integer
Dim node As Long
Dim cust As CustRecord

    If lines > MAX_LINES Then Exit Function
    If depth < Header.Height Then
        ' We are in an internal node.
        node = CachedBuckets(depth).Child(0)
        CachedIndexes(depth + 1) = node
        GetBucket CachedBuckets(depth + 1), node
        txt = txt & TreeStructure(spaces + 2, depth + 1, lines, show_leaves)
        If lines > MAX_LINES Then
            TreeStructure = txt
            Exit Function
        End If
        
        ' Handle the keys and other branches.
        For i = 1 To CachedBuckets(depth).NumKeys
            ' Display the key.
            txt = txt & Space$(spaces) & CachedBuckets(depth).Key(i) & vbCrLf
            lines = lines + 1
            If lines > MAX_LINES Then
                TreeStructure = txt & vbCrLf & "[etc.]"
                Exit Function
            End If

            ' Head down the corresponding child pointer.
            node = CachedBuckets(depth).Child(i)
            CachedIndexes(depth + 1) = node
            GetBucket CachedBuckets(depth + 1), node
            txt = txt & TreeStructure(spaces + 2, depth + 1, lines, show_leaves)
            If lines > MAX_LINES Then
                TreeStructure = txt
                Exit Function
            End If
        Next i
    Else
        ' We have reached a leaf.
        node = CachedBuckets(depth).Child(0)
        If show_leaves Then
            If node = NO_CHILD Then
                txt = txt & Space$(spaces + 2) & "<NO_CHILD>" & vbCrLf
            Else
                GetCust cust, node
                txt = txt & Space$(spaces + 2) & "<" & Trim$(cust.LastName) & "," & Trim$(cust.FirstName) & ">" & vbCrLf
            End If
        End If
        lines = lines + 1
        If lines > MAX_LINES Then
            TreeStructure = txt & vbCrLf & "[etc.]"
            Exit Function
        End If

        ' Handle the keys and other customer records.
        For i = 1 To CachedBuckets(depth).NumKeys
            ' Display the key.
            txt = txt & Space$(spaces) & CachedBuckets(depth).Key(i) & vbCrLf
            lines = lines + 1
            If lines > MAX_LINES Then
                TreeStructure = txt & vbCrLf & "[etc.]"
                Exit Function
            End If

            ' Head down the corresponding child pointer.
            node = CachedBuckets(depth).Child(i)
            If show_leaves Then
                If node = NO_CHILD Then
                    txt = txt & Space$(spaces + 2) & "<NO_CHILD>" & vbCrLf
                Else
                    GetCust cust, node
                    txt = txt & Space$(spaces + 2) & "<" & Trim$(cust.LastName) & "," & Trim$(cust.FirstName) & ">" & vbCrLf
                End If
            End If
            lines = lines + 1
            If lines > MAX_LINES Then
                TreeStructure = txt & vbCrLf & "[etc.]"
                Exit Function
            End If
        Next i
    End If
    TreeStructure = txt
End Function

' ************************************************
' Save a bucket to the disk.
' ************************************************
Private Sub WriteBucket(buck As Bucket, recnum As Long)
Dim buf As Bucket

    DiskAccesses = DiskAccesses + 1
    buf = buck
    Put #IdxFile, recnum, buf
End Sub

' ************************************************
' Save a customer record to the disk.
' ************************************************
Private Sub WriteCust(cust As CustRecord, recnum As Long)
Dim buf As CustRecord

    DiskAccesses = DiskAccesses + 1
    buf = cust
    Put #DatFile, recnum, buf
End Sub

' ************************************************
' Save the header information to the disk.
' ************************************************
Private Sub WriteHeader()
    DiskAccesses = DiskAccesses + 1
    Put #IdxFile, 1, Header
End Sub

