Attribute VB_Name = "BtreeStuff"
' ************************************************
' Btree.BAS
'
' Source code for B-tree example program.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

Global Const ORDER = 2
Global Const KEYS_PER_NODE = 2 * ORDER

' ************************************************
' Delete an item from the tree.
' ************************************************
Public Sub DeleteItem(Root As BtreeNode, txt As String)
Dim too_small As Boolean

    ' Delete the item.
    DeleteItemFromNode Root, txt, too_small
    
    ' If the root has shrunk, see if we need to
    ' remove a level from the tree.
    If Root.NumKeys < 1 Then
        Set Root = Root.Child(0)
    End If
End Sub

' ************************************************
' Recursively remove the item from somewhere down
' this node's branches.
' ************************************************
Private Sub DeleteItemFromNode(node As BtreeNode, txt As String, too_small As Boolean)
Dim match As Integer
Dim branch As Integer
Dim q As BtreeNode
Dim i As Integer

    ' If we did not find the item, say so.
    If node Is Nothing Then
        Beep
        MsgBox "Item " & txt & " is not in the tree."
        too_small = False
        Exit Sub
    End If
    
    ' Look for the item or find the branch down which
    ' it lies.
    match = False
    For branch = 1 To node.NumKeys
        If txt = node.Key(branch) Then
            match = True
            Exit For
        End If
        If txt < node.Key(branch) Then Exit For
    Next branch

    ' q is the branch down which the target lies.
    Set q = node.Child(branch - 1)
    
    If match Then
        ' We found the item. Remove it.
        If q Is Nothing Then
            ' This is a leaf node. Remove the
            ' target.
            node.NumKeys = node.NumKeys - 1
            too_small = (node.NumKeys < ORDER)
            For i = branch To node.NumKeys
                node.Key(i) = node.Key(i + 1)
            Next i
        Else
            ' Otherwise swap the item with the
            ' rightmost item to the left (in
            ' a leaf).
            SwapNode node, branch, q, too_small
            
            ' If the child node is now too small,
            ' rearrange it.
            If too_small _
                Then TooSmall _
                    node, q, branch - 1, too_small
        End If
    Else
        ' We have not found the target yet, continue
        ' down branch q.
        DeleteItemFromNode q, txt, too_small

        ' If the child node is now too small, rearrange it.
        If too_small _
            Then TooSmall _
                node, q, branch - 1, too_small
    End If
End Sub


' ************************************************
' Insert an item in a node known to have an empty
' position.
' ************************************************
Private Sub InsertInNodeWithRoom(node As BtreeNode, spot As Integer, new_key As String, new_child As BtreeNode)
Dim i As Integer

    ' Make room for the new entry.
    node.NumKeys = node.NumKeys + 1
    For i = node.NumKeys To spot + 1 Step -1
        node.Key(i) = node.Key(i - 1)
        Set node.Child(i) = node.Child(i - 1)
    Next i
    
    ' Insert the new entry.
    node.Key(spot) = new_key
    Set node.Child(spot) = new_child
End Sub

' ************************************************
' Insert an item into the tree.
' ************************************************
Public Sub InsertItem(Root As BtreeNode, txt As String)
Dim up_node As BtreeNode
Dim up_key As String
Dim had_split As Boolean
Dim old_root As BtreeNode

    ' Add the new item.
    InsertItemInNode Root, txt, up_node, up_key, had_split
    
    ' If the root node split, create a new root
    ' that contains the item that was sent up to
    ' the parent.
    If had_split Then
        Set old_root = Root
        Set Root = New BtreeNode
        Root.Key(1) = up_key
        Set Root.Child(0) = old_root
        Set Root.Child(1) = up_node
        Root.NumKeys = 1
    End If
End Sub

' ************************************************
' Recursively insert the item below this node. If
' there is a split, add the item passed up to this
' node. If this node is then full, split it.
' ************************************************
Private Sub InsertItemInNode(node As BtreeNode, new_key As String, up_node As BtreeNode, up_key As String, had_split As Boolean)
Dim branch As Integer
Dim i As Integer
                          
    If node Is Nothing Then
        ' We're at the bottom of the tree. Pass
        ' the new value back up for insertion in
        ' the node above.
        Set up_node = Nothing
        up_key = new_key
        had_split = True
        Exit Sub
    End If

    ' See which branch we should investigate.
    For branch = 0 To node.NumKeys - 1
        If node.Key(branch + 1) > new_key _
            Then Exit For
    Next branch
    
    ' Branch.
    InsertItemInNode node.Child(branch), _
        new_key, up_node, up_key, had_split

    ' If we had a split, deal with it.
    If had_split Then
        If node.NumKeys < KEYS_PER_NODE Then
            ' If there's room, just insert it.
            InsertInNodeWithRoom _
                node, branch + 1, up_key, up_node
            had_split = False
        Else
            ' The node is full. Split it.
            SplitNode _
                node, branch + 1, up_key, up_node
        End If
    End If
End Sub

' ************************************************
' Split the node plus the up_key item into two
' nodes. Variable spot indicates where in the node
' the up_key item should be placed.
'
' Return the new node through parameter
' return_node. Return the key to be passed back up
' through parameter up_key.
' ************************************************
Private Sub SplitNode(node As BtreeNode, spot As Integer, up_key As String, up_node As BtreeNode)
Dim i As Integer
Dim return_node As New BtreeNode
Dim return_key As String
Dim right_child_0 As BtreeNode

    If spot <= ORDER + 1 Then
        ' The new item should be in the left node
        ' or should be passed up.
        If spot = ORDER + 1 Then
            return_key = up_key
            Set right_child_0 = up_node
        Else
            ' Save the ORDER-th entry.
            return_key = node.Key(ORDER)
            Set right_child_0 = node.Child(ORDER)

            ' Insert the new item into the left node.
            For i = ORDER To spot + 1 Step -1
                node.Key(i) = node.Key(i - 1)
                Set node.Child(i) = node.Child(i - 1)
            Next i
            node.Key(spot) = up_key
            Set node.Child(spot) = up_node
        End If
        ' Move items into the right node.
        For i = 1 To ORDER
            return_node.Key(i) = _
                node.Key(i + ORDER)
            Set return_node.Child(i) = _
                node.Child(i + ORDER)
        Next i
    Else
        ' The new item should be in the right node.
        spot = spot - ORDER - 1
        
        return_key = node.Key(ORDER + 1)
        Set right_child_0 = node.Child(ORDER + 1)

        ' Move the first set of items into the right node.
        For i = 1 To spot - 1
            return_node.Key(i) = _
                node.Key(i + ORDER + 1)
            Set return_node.Child(i) = _
                node.Child(i + ORDER + 1)
        Next i
        
        ' Put the up_key/up_node in the right node.
        return_node.Key(spot) = up_key
        Set return_node.Child(spot) = up_node
        
        ' Put the other items in the right node.
        For i = spot + 1 To ORDER
            return_node.Key(i) = _
                node.Key(i + ORDER)
            Set return_node.Child(i) = _
                node.Child(i + ORDER)
        Next i
    End If
    
    node.NumKeys = ORDER
    return_node.NumKeys = ORDER

    Set return_node.Child(0) = right_child_0
    
    Set up_node = return_node
    up_key = return_key
End Sub

' ************************************************
' We want to remove the item in node swap_node at
' position swap_position.
'
' Recursively descend the rightmost branches of
' the tree until we come to a leaf. The node we
' are examining in the recursion is target_node.
'
' Then replace the swap item with the rightmost
' item in target_node. This will make the
' target_node smaller, possibly requiring that we
' merge it.
'
' Finally, as the recursion is unwinding, check
' nodes that are shrunk to make sure they are not
' too small.
' ************************************************
Private Sub SwapNode(swap_node As BtreeNode, swap_spot As Integer, target_node As BtreeNode, too_small As Boolean)
Dim q As BtreeNode
Dim num As Integer

    num = target_node.NumKeys
    Set q = target_node.Child(num) ' Rightmost child.
    If q Is Nothing Then
        ' We have found a leaf. Make the swap.
        swap_node.Key(swap_spot) = _
            target_node.Key(num)
        target_node.NumKeys = num - 1
        too_small = (target_node.NumKeys < ORDER)
    Else
        ' We have not yet found a leaf.
        SwapNode swap_node, swap_spot, q, too_small
        If too_small Then _
            TooSmall target_node, q, _
                target_node.NumKeys, too_small
    End If
End Sub

' ************************************************
' The node "node" is too small. It is located in
' position "spot" in parent "parent".
'
' If node has a right sibling, try to redistribute
' items between node and the right sibling. If the
' right sibling has too few items (if it has
' exactly ORDER items), merge node and the right
' sibling.
'
' If the node does not have a right sibling,
' perform the steps above for the left sibling.
'
' If two nodes are merged, see if the parent is
' now too small.
' ************************************************
Private Sub TooSmall(parent As BtreeNode, node As BtreeNode, spot As Integer, too_small As Boolean)
Dim num_in_parent As Integer
Dim num_in_sibling As Integer
Dim sibling As BtreeNode
Dim num_to_move As Integer
Dim i As Integer

    num_in_parent = parent.NumKeys
    
    ' If there is a right sibling...
    If spot < num_in_parent Then
        ' Look at the right sibling.
        spot = spot + 1 ' Now points to the sibling.
        Set sibling = parent.Child(spot)
        num_in_sibling = sibling.NumKeys
        num_to_move = (num_in_sibling - ORDER + 1) \ 2
        
        ' Put item from parent at end of items in
        ' the small node. Right now it has
        ' ORDER - 1 keys.
        node.Key(ORDER) = parent.Key(spot)
        Set node.Child(ORDER) = sibling.Child(0)
        
        If num_to_move > 0 Then
            ' If we have enough items to
            ' redistribute, do so. Copy items from
            ' sibling to node.
            For i = 1 To num_to_move - 1
                node.Key(i + ORDER) = _
                    sibling.Key(i)
                node.Child(i + ORDER) = _
                    sibling.Child(i)
            Next i
            
            ' Prepare parent
            parent.Key(spot) = _
                sibling.Key(num_to_move)
            Set parent.Child(spot) = sibling
            
            ' Fill the hole in the sibling
            Set sibling.Child(0) = _
                sibling.Child(num_to_move)
            num_in_sibling = _
                num_in_sibling - num_to_move
            For i = 1 To num_in_sibling
                sibling.Key(i) = _
                    sibling.Key(i + num_to_move)
                Set sibling.Child(i) = _
                    sibling.Child(i + num_to_move)
            Next i

            ' Update the key counts.
            sibling.NumKeys = num_in_sibling
            node.NumKeys = ORDER - 1 + num_to_move
            too_small = False
        Else ' End if enough items to redistribute...
            ' If there are not enough items to
            ' redistribute, merge the nodes.
            
            ' Move entries from sibling to node. It must
            ' contain ORDER items.
            For i = 1 To ORDER
                node.Key(i + ORDER) = _
                    sibling.Key(i)
                Set node.Child(i + ORDER) = _
                    sibling.Child(i)
            Next i
            
            ' Fill in the hole in the parent.
            For i = spot To num_in_parent - 1
                parent.Key(i) = parent.Key(i + 1)
                Set parent.Child(i) = _
                    parent.Child(i + 1)
            Next i

            ' Update the key counts.
            node.NumKeys = KEYS_PER_NODE
            parent.NumKeys = num_in_parent - 1
        
            ' Free the sibling node.
            too_small = (parent.NumKeys < ORDER)
        End If  ' End if (can redistribute) ... else (merge)
    Else ' End looking at right sibling
        ' There is no right sibling. Look at the
        ' left sibling.
        Set sibling = parent.Child(spot - 1)
        num_in_sibling = sibling.NumKeys + 1
        num_to_move = (num_in_sibling - ORDER) \ 2
        
        If num_to_move > 0 Then
            ' If we have enough items to
            ' redistribute, do so.
            
            ' Make room for the new items in
            ' the node.
            For i = ORDER - 1 To 1 Step -1
                node.Key(i + num_to_move) = _
                    node.Key(i)
                Set node.Child(i + num_to_move) = _
                    node.Child(i)
            Next i

            ' Move an item from the parent into
            ' the node.
            node.Key(num_to_move) = _
                parent.Key(spot)
            Set node.Child(num_to_move) = _
                node.Child(0)
            
            ' Move items from the sibling into
            ' the node.
            num_in_sibling = num_in_sibling - num_to_move
            For i = num_to_move - 1 To 1 Step -1
                node.Key(i) = _
                    sibling.Key(i + num_in_sibling)
                Set node.Child(i) = _
                    sibling.Child(i + num_in_sibling)
            Next i
            Set node.Child(0) = _
                sibling.Child(num_in_sibling)

            ' Revise the sibling's entry in the parent.
            parent.Key(spot) = _
                sibling.Key(num_in_sibling)

            ' Update the key counts.
            sibling.NumKeys = num_in_sibling - 1
            node.NumKeys = ORDER - 1 + num_to_move
            too_small = False
        Else ' End if enough items to redistribute
            ' Not enough items to redistribute.
            ' Merge the nodes.
            
            ' Put the node's parent entry in the sibling.
            sibling.Key(num_in_sibling) = _
                parent.Key(spot)
            Set sibling.Child(num_in_sibling) = _
                node.Child(0)
            
            ' Move items from the node into the sibling.
            For i = 1 To ORDER - 1
                sibling.Key(i + num_in_sibling) = _
                    node.Key(i)
                Set sibling.Child(i + num_in_sibling) = _
                    node.Child(i)
            Next i

            ' Update the key counts.
            sibling.NumKeys = KEYS_PER_NODE
            parent.NumKeys = num_in_parent - 1

            too_small = (parent.NumKeys < ORDER)
        End If  ' End if (can redistribute) ... else (merge)
    End If ' End if (right sibling) ... else (left sibling)
End Sub

