Attribute VB_Name = "BandBStuff"
' ************************************************
' BandB.BAS
'
' Source code for branch and bound program.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

Private unassigned_profit As Integer ' Total of unassigned profits

Public NumItems As Integer
Public MaxItem As Integer

Global Const OPTION_EXHAUSTIVE_SEARCH = 0
Global Const OPTION_BRANCH_AND_BOUND = 1

Type Item
    Cost As Integer
    Profit As Integer
End Type

Global Items() As Item
Global NodesVisited As Long
Global ToSpend As Integer
Global best_cost As Integer
Global best_profit As Integer

' True for the items in the best solution so far.
Public best_solution() As Boolean

' Solution we are testing.
Private test_solution() As Boolean
Private test_cost As Integer
Private test_profit As Integer

' ************************************************
' Perform a branch and bound search starting with
' this item.
' ************************************************
Public Sub BranchAndBound(item_num As Integer)
Dim i As Integer

    NodesVisited = NodesVisited + 1
    
    ' If this is a leaf node, it must be a better
    ' solution than we have so far or it would
    ' have been cut off earlier in the search.
    If item_num > MaxItem Then
        For i = 0 To MaxItem
            best_solution(i) = test_solution(i)
            best_profit = test_profit
            best_cost = test_cost
        Next i
        Exit Sub
    End If

    ' Otherwise descend down the child branches.
    ' First try including this item. Make sure it
    ' fits within the cost bound.
    If test_cost + Items(item_num).Cost <= ToSpend Then
        ' Add the item to the test solution.
        test_solution(item_num) = True
        test_cost = test_cost + Items(item_num).Cost
        test_profit = test_profit + Items(item_num).Profit
        unassigned_profit = unassigned_profit - Items(item_num).Profit
        
        ' Recursively see what the result might be.
        BranchAndBound item_num + 1
        
        ' Remove the item from the test solution.
        test_solution(item_num) = False
        test_cost = test_cost - Items(item_num).Cost
        test_profit = test_profit - Items(item_num).Profit
        unassigned_profit = unassigned_profit + Items(item_num).Profit
    End If

    ' Now try excluding the item. See if the
    ' remaining items have enough profit to make
    ' a path down this branch reach our lower
    ' bound.
    unassigned_profit = unassigned_profit - Items(item_num).Profit
    If test_profit + unassigned_profit > best_profit Then BranchAndBound item_num + 1
    unassigned_profit = unassigned_profit + Items(item_num).Profit
End Sub

' ************************************************
' Exhaustively search the tree for the best
' solution starting with this item.
' ************************************************
Public Sub ExhaustiveSearch(item_num As Integer)
Dim i As Integer

    NodesVisited = NodesVisited + 1

    If item_num > MaxItem Then
        ' If this is a leaf node, evaluate our success.
        If test_cost <= ToSpend And test_profit > best_profit Then
            For i = 0 To MaxItem
                best_solution(i) = test_solution(i)
                best_profit = test_profit
                best_cost = test_cost
            Next i
        End If
    Else
        ' Not a leaf node. Recursively descend.
        ' First try including this item.
        test_solution(item_num) = True
        test_cost = test_cost + Items(item_num).Cost
        test_profit = test_profit + Items(item_num).Profit
        ExhaustiveSearch item_num + 1

        ' Now try excluding the item.
        test_solution(item_num) = False
        test_cost = test_cost - Items(item_num).Cost
        test_profit = test_profit - Items(item_num).Profit
        ExhaustiveSearch item_num + 1
    End If
End Sub

' ************************************************
' Initialize variables and begin the search.
' ************************************************
Public Sub Search(search_type As Integer)
Dim i As Integer

    ' Dimension the solution arrays.
    ReDim best_solution(0 To MaxItem)
    ReDim test_solution(0 To MaxItem)

    ' Initialize as if we are investing in nothing.
    NodesVisited = 0
    best_profit = 0
    best_cost = 0
    unassigned_profit = 0
    For i = 0 To MaxItem
        unassigned_profit = unassigned_profit + Items(i).Profit
    Next i
    test_profit = 0
    test_cost = 0
    
    ' Start the search with the first item.
    Select Case search_type
        Case OPTION_EXHAUSTIVE_SEARCH
            ExhaustiveSearch 0
        
        Case OPTION_BRANCH_AND_BOUND
            BranchAndBound 0

    End Select
End Sub

