Attribute VB_Name = "Heuristics"
' ************************************************
' Heur.BAS
'
' Source code for heurstic example program.
' ************************************************
' Copyright (C) 1997 John Wiley & Sons, Inc.
' All rights reserved. See additional copyright
' information in RIGHTS.TXT.
' ************************************************
Option Explicit

Dim unassigned_profit As Integer ' Total of unassigned values

Type Item
    Cost As Integer
    Profit As Integer
End Type

Global Items() As Item
Global ToSpend As Integer
Global best_cost As Integer
Global best_profit As Integer

' best_solution is an array of used flags
' corresponding to the best solution so far.
Global best_solution() As Boolean

' Solution we are testing.
Global test_solution() As Boolean
Global test_cost As Integer
Global test_profit As Integer

' Trial solutions used by incremental change
' algorithms.
Global trial_solution() As Boolean
Global trial_cost As Integer
Global trial_profit As Integer

' The index of the last item actually in use.
Global NumItems As Integer

' ************************************************
' Add a random item to the test solution. Return
' True if an item was added to the solution.
' Return False if no more items will fit in the
' solution.
' ************************************************
Private Function AddToSolution() As Boolean
Dim num_left As Integer
Dim j As Integer
Dim selection As Integer

    ' See how many items remain that will fit within
    ' the cost bound.
    num_left = 0
    For j = 1 To NumItems
        If (Not test_solution(j)) And (test_cost + Items(j).Cost <= ToSpend) Then num_left = num_left + 1
    Next j
    
    ' Stop when we cannot find another item that fits.
    If num_left < 1 Then
        AddToSolution = False
        Exit Function
    End If
    
    ' Pick one at random.
    selection = Int((num_left) * Rnd + 1)
    
    ' Find the randomly chosen item.
    For j = 1 To NumItems
        If (Not test_solution(j)) And (test_cost + Items(j).Cost <= ToSpend) Then
            selection = selection - 1
            If selection < 1 Then Exit For
        End If
    Next j

    test_profit = test_profit + Items(j).Profit
    test_cost = test_cost + Items(j).Cost
    test_solution(j) = True

    AddToSolution = True
End Function

' ************************************************
' Switch K items at a time to improve the test
' solution.
'
' Repeat trials until we have max_bad_trials runs
' in a row without an improvement.
'
' During each trial try random changes until we
' try max_non_changes in a row without an
' improvement.
' ************************************************
Public Sub AnnealTrial(K As Integer, max_non_changes As Integer, max_back_slips As Integer)
Const TFACTOR = 0.95

Dim i As Integer
Dim non_changes As Integer
Dim t As Double
Dim max_profit As Integer
Dim min_profit As Integer
Dim doit As Boolean
Dim back_slips As Integer

    ' See what items have the largest and
    ' smallest profits.
    max_profit = Items(1).Profit
    min_profit = max_profit
    For i = 2 To NumItems
        If max_profit < Items(i).Profit Then max_profit = Items(i).Profit
        If min_profit > Items(i).Profit Then min_profit = Items(i).Profit
    Next i
    
    t = 0.75 * (max_profit - min_profit)
    back_slips = 0
    
    ' Find a random test solution to use as a
    ' starting point.
    Do While AddToSolution()
        ' All the work is done by AddToSolution.
    Loop

    ' Start with this as the trial solution.
    best_profit = test_profit
    best_cost = test_cost
    For i = 1 To NumItems
        best_solution(i) = test_solution(i)
    Next i

    ' Repeat until we try max_non_changes in a row
    ' without an improvement.
    non_changes = 0
    Do While non_changes < max_non_changes
        ' Remove a random item.
        For i = 1 To K
            RemoveFromSolution
        Next i
        
        ' Add back as many random items as will fit.
        Do While AddToSolution()
            ' All the work is done by AddToSolution.
        Loop
    
        ' If this improves the trial, save it.
        ' Otherwise reset the trial to the previous
        ' value.
        If test_profit > best_profit Then
            doit = True
        ElseIf test_profit < best_profit Then
            doit = (Rnd < Exp((test_profit - best_profit) / t))
            back_slips = back_slips + 1
            If back_slips > max_back_slips Then
                back_slips = 0
                t = t * TFACTOR
            End If
        Else
            doit = False
        End If
        If doit Then
            ' Save the improvement.
            best_profit = test_profit
            best_cost = test_cost
            For i = 1 To NumItems
                best_solution(i) = test_solution(i)
            Next i
            non_changes = 0 ' This was a good change.
        Else
            ' Reset the trial.
            test_profit = best_profit
            test_cost = best_cost
            For i = 1 To NumItems
                test_solution(i) = best_solution(i)
            Next i
            non_changes = non_changes + 1 ' Bad change.
        End If
    Loop    ' Continue trying random changes.
End Sub

' ************************************************
' Perform a balanced profit search.
' ************************************************
Public Sub BalancedProfit()
Dim i As Integer
Dim j As Integer
Dim good_ratio As Double
Dim good_j As Integer
    
    ' Repeatedly pass through the list looking for the
    ' remaining item with the largest profit/cost ratio
    ' that will fit within the cost bounds.
    For i = 1 To NumItems
        good_ratio = 0#
        good_j = -1
        ' Find the unused item with the largest
        ' profit/cost ratio that will still fit in
        ' the solution.
        For j = 1 To NumItems
            ' Make sure it is not already in the solution.
            If (Not test_solution(j)) And (test_cost + Items(j).Cost <= ToSpend) And (good_ratio < Items(j).Profit / CDbl(Items(j).Cost)) Then
                good_ratio = Items(j).Profit / CDbl(Items(j).Cost)
                good_j = j
            End If
        Next j
        
        ' Stop when we cannot find another item that fits.
        If good_j < 0 Then Exit For
        
        test_cost = test_cost + Items(good_j).Cost
        test_solution(good_j) = True
        test_profit = test_profit + Items(good_j).Profit
    Next i
End Sub

' ************************************************
' Perform a branch and bound search starting with
' this item.
' ************************************************
Public Sub BranchAndBound(Item As Integer)
Dim i As Integer

    ' If this is a leaf node, it must be a better solution
    ' than we have so far or it would have been cut off
    ' earlier in the search.
    If Item > NumItems Then
        For i = 1 To NumItems
            best_solution(i) = test_solution(i)
            best_profit = test_profit
            best_cost = test_cost
        Next i
        Exit Sub
    End If

    ' Otherwise descend down the child branches.
    
    ' First try including this item. Make sure it fits
    ' within the cost bound.
    If test_cost + Items(Item).Cost <= ToSpend Then
        test_solution(Item) = True
        test_cost = test_cost + Items(Item).Cost
        test_profit = test_profit + Items(Item).Profit
        unassigned_profit = unassigned_profit - Items(Item).Profit
        
        BranchAndBound Item + 1
        
        test_solution(Item) = False
        test_cost = test_cost - Items(Item).Cost
        test_profit = test_profit - Items(Item).Profit
        unassigned_profit = unassigned_profit + Items(Item).Profit
    End If

    ' Now try excluding the item. See if the remaining
    ' items have enough profit to make a path down this
    ' branch reach our lower bound.
    unassigned_profit = unassigned_profit - Items(Item).Profit
    If test_profit + unassigned_profit > best_profit Then BranchAndBound Item + 1
    unassigned_profit = unassigned_profit + Items(Item).Profit
End Sub

' ************************************************
' Reset the test_solution and best_solution for
' the next method.
' ************************************************
Public Sub ClearSolutions()
Dim i As Integer

    best_profit = 0
    best_cost = 0
    test_profit = 0
    test_cost = 0
    For i = 1 To NumItems
        best_solution(i) = False
        test_solution(i) = False
    Next i

    unassigned_profit = 0
    For i = 1 To NumItems
        unassigned_profit = unassigned_profit + Items(i).Profit
    Next i
End Sub

' ************************************************
' Exhaustively search the tree for the best
' solution starting with this item.
' ************************************************
Public Sub ExhaustiveSearch(Item As Integer)
Dim i As Integer

    If Item > NumItems Then
        ' If this is a leaf node, evaluate our success.
        If test_cost <= ToSpend And test_profit > best_profit Then
            For i = 1 To NumItems
                best_solution(i) = test_solution(i)
                best_profit = test_profit
                best_cost = test_cost
            Next i
        End If
    Else
        ' Not a leaf node. Recursively descend.
        ' First try including this item.
        test_solution(Item) = True
        test_cost = test_cost + Items(Item).Cost
        test_profit = test_profit + Items(Item).Profit
        ExhaustiveSearch Item + 1

        ' Now try excluding the item.
        test_solution(Item) = False
        test_cost = test_cost - Items(Item).Cost
        test_profit = test_profit - Items(Item).Profit
        ExhaustiveSearch Item + 1
    End If
End Sub

' ************************************************
' Use a hill climbing heuristic to find a
' solution.
' ************************************************
Public Sub HillClimbing()
Dim i As Integer
Dim j As Integer
Dim big_value As Integer
Dim big_j As Integer

    ' Repeatedly pass through the list looking for
    ' the remaining item with the largest profit
    ' that will fit within the cost bounds.
    For i = 1 To NumItems
        big_value = 0
        big_j = -1
        For j = 1 To NumItems
            ' Make sure it is not already in the
            ' solution.
            If (Not test_solution(j)) And (test_cost + Items(j).Cost <= ToSpend) And (big_value < Items(j).Profit) Then
                big_value = Items(j).Profit
                big_j = j
            End If
        Next j
        
        ' Stop when we cannot find another item
        ' that fits.
        If big_j < 0 Then Exit For
        
        test_cost = test_cost + Items(big_j).Cost
        test_solution(big_j) = True
        test_profit = test_profit + Items(big_j).Profit
    Next i
End Sub

' ************************************************
' Use a least cost heuristic to find a solution.
' ************************************************
Public Sub LeastCost()
Dim i As Integer
Dim j As Integer
Dim small_cost As Integer
Dim small_j As Integer

    ' Repeatedly pass through the list looking for the
    ' remaining item with the least cost that will fit
    ' within the cost bounds.
    For i = 1 To NumItems
        small_cost = 32000
        small_j = -1
        For j = 1 To NumItems
            ' Make sure it is not already in the solution.
            If (Not test_solution(j)) And (test_cost + Items(j).Cost <= ToSpend) And (small_cost > Items(j).Cost) Then
                small_cost = Items(j).Cost
                small_j = j
            End If
        Next j
        
        ' Stop when we cannot find another item that fits.
        If small_j < 0 Then Exit For
        
        test_cost = test_cost + Items(small_j).Cost
        test_solution(small_j) = True
        test_profit = test_profit + Items(small_j).Profit
    Next i
End Sub

' ************************************************
' Switch K items at a time to improve the test
' solution.
'
' Run num_trials trials making num_changes changes
' during each.
' ************************************************
Public Sub MakeChangesFixed(K As Integer, num_trials As Integer, num_changes As Integer)
Dim trial As Integer
Dim change As Integer
Dim i As Integer
Dim removal As Integer
                        
    For trial = 1 To num_trials
        ' Find a random test solution to use as a
        ' starting point.
        Do While AddToSolution()
            ' All the work is done by AddToSolution.
        Loop

        ' Start with this as the trial solution.
        trial_profit = test_profit
        trial_cost = test_cost
        For i = 1 To NumItems
            trial_solution(i) = test_solution(i)
        Next i

        For change = 1 To num_changes
            ' Remove K random items.
            For removal = 1 To K
                RemoveFromSolution
            Next removal

            ' Add back as many random items as
            ' will fit.
            Do While AddToSolution()
                ' All the work is done by AddToSolution.
            Loop
        
            ' If this improves the trial, save it.
            ' Otherwise reset the trial to the
            ' previous value.
            If test_profit > trial_profit Then
                ' Save the improvement.
                trial_profit = test_profit
                trial_cost = test_cost
                For i = 1 To NumItems
                    trial_solution(i) = test_solution(i)
                Next i
            Else
                ' Reset the trial.
                test_profit = trial_profit
                test_cost = trial_cost
                For i = 1 To NumItems
                    test_solution(i) = trial_solution(i)
                Next i
            End If
        Next change

        ' If this trial is better than the best
        ' solution so far, save it.
        If trial_profit > best_profit Then
            best_profit = trial_profit
            best_cost = trial_cost
            For i = 1 To NumItems
                best_solution(i) = trial_solution(i)
            Next i
        End If

        ' Reset the test solution for the
        ' next trial.
        test_profit = 0
        test_cost = 0
        For i = 1 To NumItems
            test_solution(i) = False
        Next i
    Next trial
End Sub

' ************************************************
' Switch K items at a time to improve the test
' solution.
'
' Repeat trials until we have max_bad_trials runs
' in a row without an improvement.
'
' During each trial try random changes until we
' try max_non_changes in a row without an
' improvement.
' ************************************************
Public Sub MakeChangesNoChange(K As Integer, max_bad_trials As Integer, max_non_changes As Integer)
Dim i As Integer
Dim removal As Integer
Dim bad_trials As Integer   ' # consecutive ineffective trials.
Dim non_changes As Integer  ' # consecutive ineffective-changes.
                        
    ' Repeat trials until we have max_bad_trials
    ' runs in a row without an improvement.
    bad_trials = 0
    Do
        ' Find a random test solution to use as a
        ' starting point.
        Do While AddToSolution()
            ' All the work is done by AddToSolution.
        Loop

        ' Start with this as the trial solution.
        trial_profit = test_profit
        trial_cost = test_cost
        For i = 1 To NumItems
            trial_solution(i) = test_solution(i)
        Next i

        ' Repeat until we try max_non_changes in
        ' a row without an improvement.
        non_changes = 0
        Do While non_changes < max_non_changes
            ' Remove K random items.
            For removal = 1 To K
                RemoveFromSolution
            Next removal

            ' Add back as many random items as will fit.
            Do While AddToSolution()
                ' All the work is done by
                ' AddToSolution.
            Loop
        
            ' If this improves the trial, save it.
            ' Otherwise reset the trial to the
            ' previous value.
            If test_profit > trial_profit Then
                ' Save the improvement.
                trial_profit = test_profit
                trial_cost = test_cost
                For i = 1 To NumItems
                    trial_solution(i) = test_solution(i)
                Next i
                non_changes = 0 ' This was a good change.
            Else
                ' Reset the trial.
                test_profit = trial_profit
                test_cost = trial_cost
                For i = 1 To NumItems
                    test_solution(i) = trial_solution(i)
                Next i
                non_changes = non_changes + 1   ' Bad change.
            End If
        Loop    ' Continue trying random changes.

        ' If this trial is better than the best
        ' solution so far, save it.
        If trial_profit > best_profit Then
            best_profit = trial_profit
            best_cost = trial_cost
            For i = 1 To NumItems
                best_solution(i) = trial_solution(i)
            Next i
            bad_trials = 0 ' This was a good trial.
        Else
            bad_trials = bad_trials + 1 ' Bad trial.
        End If

        ' Reset the test solution for the
        ' next trial.
        test_profit = 0
        test_cost = 0
        For i = 1 To NumItems
            test_solution(i) = False
        Next i
    Loop While bad_trials < max_bad_trials
End Sub

' ************************************************
' Use a random heuristic to find a solution.
' ************************************************
Public Sub RandomSearch()
Dim num_trials As Integer
Dim trial As Integer
Dim i As Integer

    ' Make several trials and keep the best.
    num_trials = NumItems   ' Use N trials.
    For trial = 1 To num_trials
        ' Make random selections until no more items fit.
        Do While AddToSolution()
            ' All the work is done by AddToSolution.
        Loop
    
        ' See if this solution is an improvement.
        If test_profit > best_profit Then
            best_profit = test_profit
            best_cost = test_cost
            For i = 1 To NumItems
                best_solution(i) = test_solution(i)
            Next i
        End If

        ' Reset the test solution for the next trial.
        test_profit = 0
        test_cost = 0
        For i = 1 To NumItems
            test_solution(i) = False
        Next i
    Next trial
End Sub

' ************************************************
' Remove a random item from the test solution.
' ************************************************
Private Sub RemoveFromSolution()
Dim num_in_solution As Integer
Dim j As Integer
Dim selection As Integer

    ' See how many items are in the solution.
    num_in_solution = 0
    For j = 1 To NumItems
        If test_solution(j) Then num_in_solution = num_in_solution + 1
    Next j
    If num_in_solution < 1 Then Exit Sub
    
    ' Pick one at random.
    selection = Int((num_in_solution) * Rnd + 1)
    
    ' Find the randomly chosen item.
    For j = 1 To NumItems
        If test_solution(j) Then
            selection = selection - 1
            If selection < 1 Then Exit For
        End If
    Next j

    ' Remove the item from the solution.
    test_profit = test_profit - Items(j).Profit
    test_cost = test_cost - Items(j).Cost
    test_solution(j) = False
End Sub

